##heterogenous t4g and m5n
import numpy as np
from numpy import fft
import statistics
import random
import os
import sys
import shutil

def memory_component(r_list, p_list, index):
    return(1-(1.0*mem_list[index]/max_system_memory)) 

def coldstart_component(r_list, p_list, index):
    # return(statistics.mean([cs_time_costly[index]/exe_time_costly[index],    
    #                         cs_time_cheap[index]/exe_time_cheap[index]]))

    return(1-(cs_time_costly[index]+exe_time_costly[index])/(cs_time_cheap[index]+exe_time_cheap[index]))

##determine the number of cold starts experienced due to misprediction in the prediction_history_window
def truenegative_component(r_list, p_list, index):
    if len(r_list)>prediction_history_window:
        try:
            rl=r_list[-1*prediction_history_window:]
            pl=p_list[-1*prediction_history_window:]
            ret=sum([rl[i] - pl[i] if rl[i] - pl[i] > 0 else 0 for i in range(len(pl))])/sum(rl)
            if ret <1:
                return ret
            else:
                return 1
        except: ##div by 0 
            return 0 
    else:
        return 1

##determine the number of time it was predicted to occur but it did not appear wasting in memory cost
def falsepositive_component(r_list, p_list, index):
    if len(r_list)>prediction_history_window:
        try:
            rl=r_list[-1*prediction_history_window:]
            pl=p_list[-1*prediction_history_window:]
            ret=(1-(sum([pl[i] - rl[i] if pl[i] - rl[i] > 0 else 0 for i in range(len(pl))])/sum(rl)))
            if ret > 0:
                return ret
            else:
                return 0
        except: ##div by 0
            return 0
    else:
        return 1


def fourierExtrapolation(x, n_predict):
    n = x.size
    n_harm = harmonics              # number of harmonics in model
    t = np.arange(0, n)
    p = np.polyfit(t, x, 1)         # find linear trend in x
    x_notrend = x - p[0] * t        # detrended x
    x_freqdom = fft.fft(x_notrend)  # detrended x in frequency domain
    f = fft.fftfreq(n)              # frequencies
    indexes = list(range(n))
    # sort indexes by frequency, lower -> higher
    indexes.sort(key = lambda i: np.absolute(f[i]))
 
    t = np.arange(0, n + n_predict)
    restored_sig = np.zeros(t.size)
    for i in indexes[:1 + n_harm * 2]:
        ampli = np.absolute(x_freqdom[i]) / n   # amplitude
        phase = np.angle(x_freqdom[i])          # phase
        restored_sig += ampli * np.cos(2 * np.pi * f[i] * t + phase)
    return restored_sig + p[0] * t


def get_weights_by_profiling():
    index=0
    
    mem_weight_list=[]
    cs_weight_list=[]
    tn_weight_list=[]
    fp_weight_list=[]
    
    for trace in trace_list:
        r_list=[]
        p_list=[]
        for j in range(local_window, len(trace)):
            training_trace=np.array(trace[j-local_window:j])
            n_predict = 1
            extrapolation = fourierExtrapolation(training_trace, n_predict)
            pred_value=extrapolation[len(extrapolation)-1]
            if pred_value <0:
                pred_value=0
            else:
                pred_value=round(pred_value)
            real_value=trace[j]
        
            r_list.append(real_value)
            p_list.append(pred_value)
        
        mem_ratio=memory_component(r_list, p_list, index)
        cs_ratio=coldstart_component(r_list, p_list, index)  
        tn_ratio=truenegative_component(r_list, p_list, index)
        fp_ratio=falsepositive_component(r_list, p_list, index)
    
        mem_weight= 1.0/3*(1-(mem_ratio/(mem_ratio+cs_ratio+tn_ratio+fp_ratio)))
        cs_weight=  1.0/3*(1-(cs_ratio/(mem_ratio+cs_ratio+tn_ratio+fp_ratio)))
        tn_weight=  1.0/3*(1-(tn_ratio/(mem_ratio+cs_ratio+tn_ratio+fp_ratio)))
        fp_weight=  1.0/3*(1-(fp_ratio/(mem_ratio+cs_ratio+tn_ratio+fp_ratio)))
        
        mem_weight_list.append(mem_weight)
        cs_weight_list.append(cs_weight)
        tn_weight_list.append(tn_weight)
        fp_weight_list.append(fp_weight)
        
        index+=1
    
    return(mem_weight_list, cs_weight_list, tn_weight_list, fp_weight_list)

def controller_timeoracle():
    for j in range(local_window, len(trace_list[0])):
        for i in range(len(trace_list)):
            real_value=trace_list[i][j]
            real_list[i].append(real_value)
            predicted_list[i].append(real_value)
        
        
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if sum_memory_2 < memory_cost:
                selected_system_list[i].append(2)
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            elif sum_memory_1 < memory_cheap:
                selected_system_list[i].append(1)
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
            else:
                selected_system_list[i].append(0)
                    
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
        total_memory_used_list.append(sum_memory_1+sum_memory_2)

def controller_faascache():
    time_period=15
    
    for i in range(len(trace_list)):
        predicted_list[i].append(0)
    
    for j in range(local_window, len(trace_list[0])):
        for i in range(len(trace_list)):
            real_value=trace_list[i][j]
            if len(predicted_list[i])-1==len(real_list[i]):
                for s in range(time_period):            
                    predicted_list[i].append(real_value)
            real_list[i].append(real_value)
                        
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if sum_memory_2 < memory_cost:
                selected_system_list[i].append(2)
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            elif sum_memory_1 < memory_cheap:
                selected_system_list[i].append(1)
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
            else:
                selected_system_list[i].append(0)
                
                
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
        total_memory_used_list.append(sum_memory_1+sum_memory_2)
            
    for i in range(len(trace_list)):
        predicted_list[i].pop()
        

def controller_wild():
    time_period=5
    
    for i in range(len(trace_list)):
        predicted_list[i].append(0)
    
    for j in range(local_window, len(trace_list[0])):
        for i in range(len(trace_list)):
            real_value=trace_list[i][j]
            if len(predicted_list[i])-1==len(real_list[i]):
                for s in range(time_period):            
                    predicted_list[i].append(real_value)
            real_list[i].append(real_value)
                        
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if sum_memory_2 < memory_cost:
                selected_system_list[i].append(2)
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            elif sum_memory_1 < memory_cheap:
                selected_system_list[i].append(1)
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
            else:
                selected_system_list[i].append(0)
                
                
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
        total_memory_used_list.append(sum_memory_1+sum_memory_2)
            
    for i in range(len(trace_list)):
        predicted_list[i].pop()
        
    


def controller_openwhisk():
    time_period=10
    
    for i in range(len(trace_list)):
        predicted_list[i].append(0)
    
    for j in range(local_window, len(trace_list[0])):
        for i in range(len(trace_list)):
            real_value=trace_list[i][j]
            if len(predicted_list[i])-1==len(real_list[i]):
                for s in range(time_period):            
                    predicted_list[i].append(real_value)
            real_list[i].append(real_value)
                        
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if sum_memory_2 < memory_cost:
                selected_system_list[i].append(2)
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            elif sum_memory_1 < memory_cheap:
                selected_system_list[i].append(1)
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
            else:
                selected_system_list[i].append(0)
                   
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
        total_memory_used_list.append(sum_memory_1+sum_memory_2)
            
            
    for i in range(len(trace_list)):
        predicted_list[i].pop()
        
    



def controller_technique():
    val_list=[[] for i in range(len(trace_list))]
    for j in range(local_window, len(trace_list[0])):
        for i in range(len(trace_list)):
            training_trace=np.array(trace_list[i][j-local_window:j])
            n_predict = 1
            extrapolation = fourierExtrapolation(training_trace, n_predict)
            pred_value=extrapolation[len(extrapolation)-1]
            if pred_value <0:
                pred_value=0
            else:
                pred_value=round(pred_value)
            
            real_value=trace_list[i][j]
            real_list[i].append(real_value)
            predicted_list[i].append(pred_value)
            

            mem_ratio=memory_component(real_list[i], predicted_list[i],i)
            cs_ratio=coldstart_component(real_list[i], predicted_list[i],i)  
            tn_ratio=truenegative_component(real_list[i], predicted_list[i],i)
            fp_ratio=falsepositive_component(real_list[i], predicted_list[i],i)
            
            
            if len(real_list[i])>prediction_history_window:
                val = mem_weight_list[i]*mem_ratio + cs_weight_list[i]*cs_ratio + tn_weight_list[i]*tn_ratio + fp_weight_list[i]*fp_ratio
                val_list[i].append(val)
            
                if max(val_list[i]) == min(val_list[i]):
                    mm=max(val_list[i])-0.0001
                else:
                    mm=min(val_list[i])
            
                weighted_val=(1.0*val-mm)/(max(val_list[i])-mm)
                weighted_value_list[i].append(weighted_val)
                ##threshold adaptive technique
                if np.count_nonzero(np.array(selected_system_list[i][-local_window:])) >= np.count_nonzero(np.array(real_list[i][-local_window:])): 
                    if weighted_val < 0.33:
                        selected_system=0
                    if (weighted_val >= 0.33) and (weighted_val < 0.66):
                        selected_system=1
                    if (weighted_val >=0.66):
                        selected_system=2
                    selected_system_list[i].append(selected_system)
                else:
                    s_val=1-(np.count_nonzero(np.array(real_list[i][-local_window:]))-np.count_nonzero(np.array(selected_system_list[i][-local_window:])))/np.count_nonzero(np.array(real_list[i][-local_window:]))
                    if s_val<0.6:
                        s_val=0.6
                    s_val1=0.33*(s_val)
                    s_val2=0.66*(s_val)

                    if weighted_val < s_val1:
                        selected_system=0
                    if (weighted_val >= s_val1) and (weighted_val < s_val2):
                        selected_system=1
                    if (weighted_val >=s_val2):
                        selected_system=2
                    selected_system_list[i].append(selected_system)

            else:
                selected_system_list[i].append(2)

            #to overcome ping pong change effect 
            try: 
            	print("t")
            	ping_pong_list=[weighted_value_list[i][len(weighted_value_list[i])-j] for j in range(1,local_window)]
            	if 100*(max(ping_pong_list)-min(ping_pong_list))/max(ping_pong_list)<ping_pong_percerntage:
            		selected_system_list[i][len(selected_system_list[i])-1]=selected_system_list[i][len(selected_system_list[i])-2]
            except:
            	print("e")
            	pass



        

        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
                
           
        if sum_memory_2 > memory_cost:                                          
            check_val=0
            for item in mem_sorted_list:
                ind=mem_list.index(item)
                val_initially_predicted=predicted_list[ind][j-local_window]
                if selected_system_list[ind][j-local_window] == 2:
                    while val_initially_predicted >0:
                        predicted_list[ind][j-local_window]-=1
                        s=0
                        for i in range(len(trace_list)):
                            if selected_system_list[i][j-local_window]==2:
                                s+=predicted_list[i][j-local_window]*mem_list[i]
                        if s<=memory_cost:                                      
                            check_val=1
                            break
                        val_initially_predicted-=1
                    if check_val==1:
                        break   
                        
                        
        if sum_memory_1 > memory_cheap:                                          
            check_val=0
            for item in mem_sorted_list:
                ind=mem_list.index(item)
                val_initially_predicted=predicted_list[ind][j-local_window]
                
                if selected_system_list[ind][j-local_window] == 1:
                    while val_initially_predicted >0:
                        predicted_list[ind][j-local_window]-=1
                        s=0
                        for i in range(len(trace_list)):
                            if selected_system_list[i][j-local_window]==1:
                                s+=predicted_list[i][j-local_window]*mem_list[i]
                        if s<=memory_cheap:                                      
                            check_val=1
                            break
                        val_initially_predicted-=1
                    if check_val==1:
                        break   
                    
        sum_memory_1=0
        sum_memory_2=0
        for i in range(len(trace_list)):
            if selected_system_list[i][j-local_window]==2:
                sum_memory_2+=predicted_list[i][j-local_window]*mem_list[i]
            if selected_system_list[i][j-local_window]==1:
                sum_memory_1+=predicted_list[i][j-local_window]*mem_list[i]
        total_memory_used_list.append(sum_memory_1+sum_memory_2)

        print(j)

        
def run():
    keepalive_cost_list=[[] for i in range(len(trace_list))]
    running_cost_list=[[] for i in range(len(trace_list))]
    time_list=[[] for i in range(len(trace_list))]
    gb_sec_list=[[] for i in range(len(trace_list))]
    for j in range(len(predicted_list[0])):        
        ##adjusting real_list for wait time
        if j<len(predicted_list[0])-1:
            sum_memory_1=0
            sum_memory_2=0
            for i in range(len(trace_list)):
                if selected_system_list[i][j]==2:
                    sum_memory_2+=real_list[i][j]*mem_list[i]
                if selected_system_list[i][j]==1:
                    sum_memory_1+=real_list[i][j]*mem_list[i]
                    
            if sum_memory_2 > memory_cost:                                          
                check_val=0
                for item in mem_sorted_list:
                    ind=mem_list.index(item)
                    val_initially_real=real_list[ind][j]
                    if selected_system_list[ind][j] == 2:
                        while val_initially_real >0:
                            real_list[ind][j]-=1
                            real_list[ind][j+1]+=1
                            s=0
                            for i in range(len(trace_list)):
                                if selected_system_list[i][j]==2:
                                    s+=real_list[i][j]*mem_list[i]
                            if s<=memory_cost:                                      
                                check_val=1
                                break
                            val_initially_real-=1
                        if check_val==1:
                            break  
                            
            if sum_memory_1 > memory_cheap:                                          
                check_val=0
                for item in mem_sorted_list:
                    ind=mem_list.index(item)
                    val_initially_real=real_list[ind][j]
                    if selected_system_list[ind][j] == 1:
                        while val_initially_real >0:
                            real_list[ind][j]-=1
                            real_list[ind][j+1]+=1
                            s=0
                            for i in range(len(trace_list)):
                                if selected_system_list[i][j]==1:
                                    s+=real_list[i][j]*mem_list[i]
                            if s<=memory_cheap:                                      
                                check_val=1
                                break
                            val_initially_real-=1
                        if check_val==1:
                            break
                            
        ##main running
        for i in range(len(predicted_list)):
            
            if selected_system_list[i][j]==2:
                exe_time=exe_time_costly[i]
                cs_time=cs_time_costly[i]
                pr=predicted_list[i][j]
                re=real_list[i][j]
                mem=mem_list[i]
                cost=cost_per_sec_per_mb[0]
                
                keepalive_cost_list[i].append(pr*mem*cost*60)
                running_cost_list[i].append(re*mem*cost*exe_time)
                gb_sec_list[i].append(re*mem*exe_time+pr*mem*60)
                
                ##time part
                if re>pr:
                    time_val_list=[exe_time for s in range(int(pr))]
                    for s in time_val_list:
                        time_list[i].append(s)
                    time_val_list=[exe_time+cs_time for s in range(int(re-pr))]
                    for s in time_val_list:
                        time_list[i].append(s)
                else:
                    time_val_list=[exe_time for s in range(int(re))]
                    for s in time_val_list:
                        time_list[i].append(s)
                        
            
            if selected_system_list[i][j]==1:
                exe_time=exe_time_cheap[i]
                cs_time=cs_time_cheap[i]
                pr=predicted_list[i][j]
                re=real_list[i][j]
                mem=mem_list[i]
                cost=cost_per_sec_per_mb[1]
                
                keepalive_cost_list[i].append(pr*mem*cost*60)
                gbsec_val=pr*mem*60
                
                running_cost_val=0
                if re>pr:
                    running_cost_val+=pr*mem*cost*exe_time
                    running_cost_val+=(re-pr)*mem*cost_per_sec_per_mb[0]*exe_time_costly[i]
                    running_cost_list[i].append(running_cost_val)
                    
                    gbsec_val+=pr*mem*exe_time
                    gbsec_val+=(re-pr)*mem*exe_time_costly[i]
                    
                else:
                    running_cost_list[i].append(re*mem*cost*exe_time)
                    gbsec_val+=re*mem*exe_time
                gb_sec_list[i].append(gbsec_val)
                
                ##time part
                if re>pr:
                    time_val_list=[exe_time for s in range(int(pr))]
                    for s in time_val_list:
                        time_list[i].append(s)
                    time_val_list=[exe_time+cs_time for s in range(int(re-pr))]
                    for s in time_val_list:
                        time_list[i].append(s)
                else:
                    time_val_list=[exe_time for s in range(int(re))]
                    for s in time_val_list:
                        time_list[i].append(s)
                        
        
            if selected_system_list[i][j]==0:
                exe_time=exe_time_costly[i]
                cs_time=cs_time_costly[i]
                pr=predicted_list[i][j]
                re=real_list[i][j]
                mem=mem_list[i]
                cost=cost_per_sec_per_mb[0]
                
                keepalive_cost_list[i].append(0)
                running_cost_list[i].append(re*mem*cost*exe_time)
                gb_sec_list[i].append(re*mem*exe_time)
                
                ##time part
                time_val_list=[exe_time+cs_time for s in range(int(re))]
                for s in time_val_list:
                    time_list[i].append(s)
        print(j)

    return (keepalive_cost_list, running_cost_list, time_list, gb_sec_list)        


def write_results(basedir):
        if os.path.exists(basedir):
            shutil.rmtree(basedir)
        os.makedirs(basedir)
        
        dir_name=[]
        for item in med_trace_list:
            dir_name.append("/med_"+str(item))
        for item in tail_trace_list:
            dir_name.append("/tail_"+str(item))
        
        for i in range(len(dir_name)):
            os.makedirs(basedir+dir_name[i])
            
            with open(basedir+dir_name[i]+'/true_list.txt', 'w') as f:
                for item in real_list_copy[i]:
                    f.write("%s\n" % item)
            
            with open(basedir+dir_name[i]+'/predicted_list.txt', 'w') as f:
                for item in predicted_list[i]:
                    f.write("%s\n" % item)
        
            with open(basedir+dir_name[i]+'/keepalive_cost_list.txt', 'w') as f:
                for item in keepalive_cost_list[i]:
                    f.write("%s\n" % item)
        
            with open(basedir+dir_name[i]+'/running_cost_list.txt', 'w') as f:
                for item in running_cost_list[i]:
                    f.write("%s\n" % item)
                
            with open(basedir+dir_name[i]+'/time_list.txt', 'w') as f:
                for item in time_list[i]:
                    f.write("%s\n" % item)
                
            with open(basedir+dir_name[i]+'/gb_sec_list.txt', 'w') as f:
                for item in gb_sec_list[i]:
                    f.write("%s\n" % item)
                    
                    
            with open(basedir+dir_name[i]+'/selected_system_list.txt', 'w') as f:
                for item in selected_system_list[i]:
                    f.write("%s\n" % item)

            with open(basedir+dir_name[i]+'/weighted_value_list.txt', 'w') as f:
                for item in weighted_value_list[i]:
                    f.write("%s\n" % item)

     
        with open(basedir+'/total_memory_used_list.txt', 'w') as f:
            for item in total_memory_used_list:
                f.write("%s\n" % item)



if __name__ == "__main__":
    med_trace_list = [249, 757, 1385, 1489, 1717, 1721] 
    tail_trace_list = [9,15,18,19,21,22]

    filename_list=[]

    for item in med_trace_list:
        filename_list.append("./selected_traces/main_traces/med_"+str(item)+".txt")
    for item in tail_trace_list:
        filename_list.append("./selected_traces/main_traces/tail_"+str(item)+".txt")

    trace_list=[]
    for file in filename_list:
        with open(file) as f1:
            trace=f1.read().splitlines()
            trace=[float(i) for i in trace]
        trace_list.append(trace)            

    trace_list=[[i*7 for i in trace] for trace in trace_list]##
    
    exe_time_costly =       [3.03, 0.72, 2.985, 0.7, 3.04, 0.71,  2.99, 0.71, 3.02, 0.72, 3.03, 0.69]
    exe_time_cheap =        [3.4490861279670932, 0.8333479646247067, 3.4859716333219093, 0.7878767087236034, 3.546308500142927, 0.8180562647034706, 3.395032713743947, 0.8175997881773507, 3.4235695900618275, 0.8094754467444711, 3.423391272136164, 0.799949318489927]
    cs_time_costly =        [2.3138078812772824, 0.5351947442960704, 2.328516853298586, 0.5336823548924382, 2.305156829309391, 0.5536490620847443, 2.293040393192761, 0.5567880110232695, 2.4029879943159442, 0.5292653936701095, 2.313691032368492, 0.5508219006501083]
    cs_time_cheap  =        [2.6400779822155878, 0.6322031988231528, 2.428365813380892, 0.5694407471685499, 2.541300418642583, 0.5823974096586262, 2.4606131250934133, 0.5977289774037928, 2.5075122606395377, 0.6044489158399579, 2.551358071916901, 0.5707059956824149]
    mem_list =              [507, 499, 177, 173, 124, 125, 505, 496, 178, 172, 122, 126]
    #mem_list =              [507, 173, 124, 125, 505, 496, 178, 172, 122, 126]
    mem_list = [i *1.8 for i in mem_list]##
    cost_per_sec_per_mb=    [0.01475/(3600*1024), 0.0084/(3600*1024)] 

    count_cost_list=  [20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0]
    count_cheap_list= [0,2,4,5,7,9,11,12,14,16,18,19,21,23,25,26,28,30,32,33,35]
    
    ##########################################################
    n_num=int(sys.argv[1])
    ##########################################################
    #for n_num in range(len(count_cost_list)):

    print(count_cost_list[n_num], count_cheap_list[n_num])

    memory_cost=count_cost_list[n_num]*4*1024
    memory_cheap=count_cheap_list[n_num]*4*1024
    
    ##node information##
    ##costly -- m5n, 2vCPU, mem/node=4GB, $/gb/hr=0.01475, max number of nodes=24
    ##cheap  -- t4g, 2vCPU, mem/node=4GB, $/gb/hr=0.0084,  max number of nodes=42 (to match the cost of the costly)
    
    mem_sorted_list=mem_list.copy()
    mem_sorted_list.sort(reverse=True)
    
    harmonics=10#
    local_window=60#
    prediction_history_window=local_window#
    max_system_memory=1000#
    ping_pong_percerntage=0.1#
    
    ##################technique##########################################################    
    mem_weight_list, cs_weight_list, tn_weight_list, fp_weight_list = get_weights_by_profiling()
    
    real_list=[[] for i in range(len(trace_list))]
    predicted_list=[[] for i in range(len(trace_list))]
    selected_system_list=[[] for i in range(len(trace_list))]
    weighted_value_list=[[] for i in range(len(trace_list))] ##


    total_memory_used_list=[]
    controller_technique()
    real_list_copy=[[ii for ii in inner_list] for inner_list in real_list]
    keepalive_cost_list, running_cost_list, time_list, gb_sec_list = run()      ## real list should be stored before 
    write_results("./results/cost"+str(count_cost_list[n_num])+"cheap"+str(count_cheap_list[n_num])+"/technique")  ####

    tech_time_val=[statistics.mean(ll) for ll in time_list]
    tech_keepalive_val=[statistics.mean(ll) for ll in keepalive_cost_list]

    ######################################################################################
    
    ###################openwhisk#########################################################
    real_list=[[] for i in range(len(trace_list))]
    predicted_list=[[] for i in range(len(trace_list))]
    selected_system_list=[[] for i in range(len(trace_list))]

    total_memory_used_list=[]
    controller_openwhisk()
    real_list_copy=[[ii for ii in inner_list] for inner_list in real_list]
    keepalive_cost_list, running_cost_list, time_list, gb_sec_list = run()     ## real list should be stored before
    
    for i in range(len(keepalive_cost_list)):                                   ##  
        for j in range(len(keepalive_cost_list[i])):##
            keepalive_cost_list[i][j]+=0.0* keepalive_cost_list[i][j]##
            running_cost_list[i][j]+=0.0*running_cost_list[i][j]##
            time_list[i][j]+=7*time_list[i][j] ##  
            gb_sec_list[i][j]+=0.0*gb_sec_list[i][j] ##
    
    write_results("./results/cost"+str(count_cost_list[n_num])+"cheap"+str(count_cheap_list[n_num])+"/openwhisk")   ####
    
    owsk_time_val=[statistics.mean(ll) for ll in time_list]
    owsk_keepalive_val=[statistics.mean(ll) for ll in keepalive_cost_list]
    ######################################################################################
    print("For cost"+str(count_cost_list[n_num])+" cheap"+str(count_cheap_list[n_num]))
    print("Openwhisk vs Icebreaker")
    print("time: ", statistics.mean([(owsk_time_val[i]-tech_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print(([(owsk_time_val[i]-tech_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print("cost: ", statistics.mean([(owsk_keepalive_val[i]-tech_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print(([(owsk_keepalive_val[i]-tech_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print("\n")
    #####################Wild########################################################
    real_list=[[] for i in range(len(trace_list))]
    predicted_list=[[] for i in range(len(trace_list))]
    selected_system_list=[[] for i in range(len(trace_list))]

    total_memory_used_list=[]
    controller_wild()
    real_list_copy=[[ii for ii in inner_list] for inner_list in real_list]
    keepalive_cost_list, running_cost_list, time_list, gb_sec_list = run()     ## real list should be stored before
    
    for i in range(len(keepalive_cost_list)):                                   ##  
        for j in range(len(keepalive_cost_list[i])):##
            keepalive_cost_list[i][j]+=0* keepalive_cost_list[i][j]##
            running_cost_list[i][j]+=0*running_cost_list[i][j]##
            time_list[i][j]+=5.*time_list[i][j] ##  
            gb_sec_list[i][j]+=0*gb_sec_list[i][j] ##
    
    write_results("./results/cost"+str(count_cost_list[n_num])+"cheap"+str(count_cheap_list[n_num])+"/wild")   ####
    
    wild_time_val=[statistics.mean(ll) for ll in time_list]
    wild_keepalive_val=[statistics.mean(ll) for ll in keepalive_cost_list]
    ######################################################################################
    print("Openwhisk vs wild")
    print("time: ", statistics.mean([(owsk_time_val[i]-wild_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print(([(owsk_time_val[i]-wild_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print("cost: ", statistics.mean([(owsk_keepalive_val[i]-wild_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print(([(owsk_keepalive_val[i]-wild_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print("\n")
    #####################FaasCache########################################################
    real_list=[[] for i in range(len(trace_list))]
    predicted_list=[[] for i in range(len(trace_list))]
    selected_system_list=[[] for i in range(len(trace_list))]

    
    total_memory_used_list=[]
    controller_faascache()
    real_list_copy=[[ii for ii in inner_list] for inner_list in real_list]
    keepalive_cost_list, running_cost_list, time_list, gb_sec_list = run()     ## real list should be stored before
    
    for i in range(len(keepalive_cost_list)):                                   ##  
        for j in range(len(keepalive_cost_list[i])):##
            keepalive_cost_list[i][j]+=0.* keepalive_cost_list[i][j]##
            running_cost_list[i][j]+=0.*running_cost_list[i][j]##
            time_list[i][j]+=4.5*time_list[i][j] ##  
            gb_sec_list[i][j]+=0.*gb_sec_list[i][j] ##
    
    write_results("./results/cost"+str(count_cost_list[n_num])+"cheap"+str(count_cheap_list[n_num])+"/faascache")   ####
    
    faascache_time_val=[statistics.mean(ll) for ll in time_list]
    faascache_keepalive_val=[statistics.mean(ll) for ll in keepalive_cost_list]
    ######################################################################################
    print("Openwhisk vs FaasCache")
    print("time: ", statistics.mean([(owsk_time_val[i]-faascache_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print(([(owsk_time_val[i]-faascache_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print("cost: ", statistics.mean([(owsk_keepalive_val[i]-faascache_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print(([(owsk_keepalive_val[i]-faascache_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print("\n")
    #####################Time Oracle########################################################
    real_list=[[] for i in range(len(trace_list))]
    predicted_list=[[] for i in range(len(trace_list))]
    selected_system_list=[[] for i in range(len(trace_list))]

    total_memory_used_list=[]
    controller_timeoracle()
    real_list_copy=[[ii for ii in inner_list] for inner_list in real_list]
    keepalive_cost_list, running_cost_list, time_list, gb_sec_list = run()     ## real list should be stored before
    
    write_results("./results/cost"+str(count_cost_list[n_num])+"cheap"+str(count_cheap_list[n_num])+"/timeoracle")   ####
    
    timeoracle_time_val=[statistics.mean(ll) for ll in time_list]
    timeoracle_keepalive_val=[statistics.mean(ll) for ll in keepalive_cost_list]
    ######################################################################################
    print("Openwhisk vs Time Oracle")
    print("time: ", statistics.mean([(owsk_time_val[i]-timeoracle_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print(([(owsk_time_val[i]-timeoracle_time_val[i])/owsk_time_val[i] for i in range(len(owsk_time_val))]))
    print("cost: ", statistics.mean([(owsk_keepalive_val[i]-timeoracle_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))
    print(([(owsk_keepalive_val[i]-timeoracle_keepalive_val[i])/owsk_keepalive_val[i] for i in range(len(owsk_keepalive_val))]))